@@ -30,55 +30,43 @@
 #include "cli.h"
 #include "uevent.h"
 #include "foreign.h"
+#include "strbuf.h"
 #include "cli_handlers.h"
 
+#define SET_REPLY_AND_LEN(__rep, __len, string_literal)			\
+	do {								\
+		*(__rep) = strdup(string_literal);			\
+		*(__len) = *(__rep) ? sizeof(string_literal) : 0;	\
+	} while (0)
+
 int
 show_paths (char ** r, int * len, struct vectors * vecs, char * style,
 	    int pretty)
 {
+	STRBUF_ON_STACK(reply);
 	int i;
 	struct path * pp;
-	char * c;
-	char * reply, * header;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	int hdr_len = 0;
 
 	get_path_layout(vecs->pathvec, 1);
 	foreign_path_layout();
 
-	reply = MALLOC(maxlen);
+	if (pretty && (hdr_len = snprint_path_header(&reply, style)) < 0)
+		return 1;
 
-	while (again) {
-		if (!reply)
+	vector_foreach_slot(vecs->pathvec, pp, i) {
+		if (snprint_path(&reply, style, pp, pretty) < 0)
 			return 1;
-
-		c = reply;
-
-		if (pretty)
-			c += snprint_path_header(c, reply + maxlen - c,
-						 style);
-		header = c;
-
-		vector_foreach_slot(vecs->pathvec, pp, i)
-			c += snprint_path(c, reply + maxlen - c,
-					  style, pp, pretty);
-
-		c += snprint_foreign_paths(c, reply + maxlen - c,
-					   style, pretty);
-
-		again = (c == reply + maxlen - 1);
-
-		REALLOC_REPLY(reply, again, maxlen);
 	}
+	if (snprint_foreign_paths(&reply, style, pretty) < 0)
+		return 1;
 
-	if (pretty && c == header) {
+	if (pretty && get_strbuf_len(&reply) == (size_t)hdr_len)
 		/* No output - clear header */
-		*reply = '\0';
-		c = reply;
-	}
+		truncate_strbuf(&reply, 0);
 
-	*r = reply;
-	*len = (int)(c - reply + 1);
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -86,28 +74,14 @@ int
 show_path (char ** r, int * len, struct vectors * vecs, struct path *pp,
 	   char * style)
 {
-	char * c;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	STRBUF_ON_STACK(reply);
 
 	get_path_layout(vecs->pathvec, 1);
-	reply = MALLOC(maxlen);
+	if (snprint_path(&reply, style, pp, 0) < 0)
+		return 1;
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 
-	while (again) {
-		if (!reply)
-			return 1;
-
-		c = reply;
-
-		c += snprint_path(c, reply + maxlen - c, style, pp, 0);
-
-		again = (c == reply + maxlen - 1);
-
-		REALLOC_REPLY(reply, again, maxlen);
-	}
-	*r = reply;
-	*len = (int)(c - reply + 1);
 	return 0;
 }
 
@@ -115,84 +89,51 @@ int
 show_map_topology (char ** r, int * len, struct multipath * mpp,
 		   struct vectors * vecs)
 {
-	char * c;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	STRBUF_ON_STACK(reply);
 
 	if (update_multipath(vecs, mpp->alias, 0))
 		return 1;
-	reply = MALLOC(maxlen);
 
-	while (again) {
-		if (!reply)
-			return 1;
+	if (snprint_multipath_topology(&reply, mpp, 2) < 0)
+		return 1;
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 
-		c = reply;
-
-		c += snprint_multipath_topology(c, reply + maxlen - c, mpp, 2);
-		again = (c == reply + maxlen - 1);
-
-		REALLOC_REPLY(reply, again, maxlen);
-	}
-	*r = reply;
-	*len = (int)(c - reply + 1);
 	return 0;
 }
 
 int
 show_maps_topology (char ** r, int * len, struct vectors * vecs)
 {
+	STRBUF_ON_STACK(reply);
 	int i;
 	struct multipath * mpp;
-	char * c;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
 
 	get_path_layout(vecs->pathvec, 0);
 	foreign_path_layout();
 
-	reply = MALLOC(maxlen);
-
-	while (again) {
-		if (!reply)
-			return 1;
-
-		c = reply;
-
-		vector_foreach_slot(vecs->mpvec, mpp, i) {
-			if (update_multipath(vecs, mpp->alias, 0)) {
-				i--;
-				continue;
-			}
-			c += snprint_multipath_topology(c, reply + maxlen - c,
-							mpp, 2);
+	vector_foreach_slot(vecs->mpvec, mpp, i) {
+		if (update_multipath(vecs, mpp->alias, 0)) {
+			i--;
+			continue;
 		}
-		c += snprint_foreign_topology(c, reply + maxlen - c, 2);
-
-		again = (c == reply + maxlen - 1);
-
-		REALLOC_REPLY(reply, again, maxlen);
+		if (snprint_multipath_topology(&reply, mpp, 2) < 0)
+			return 1;
 	}
+	if (snprint_foreign_topology(&reply, 2) < 0)
+		return 1;
 
-	*r = reply;
-	*len = (int)(c - reply + 1);
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
 int
 show_maps_json (char ** r, int * len, struct vectors * vecs)
 {
+	STRBUF_ON_STACK(reply);
 	int i;
 	struct multipath * mpp;
-	char * c;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
-
-	if (VECTOR_SIZE(vecs->mpvec) > 0)
-		maxlen *= PRINT_JSON_MULTIPLIER * VECTOR_SIZE(vecs->mpvec);
 
 	vector_foreach_slot(vecs->mpvec, mpp, i) {
 		if (update_multipath(vecs, mpp->alias, 0)) {
@@ -200,21 +141,11 @@ show_maps_json (char ** r, int * len, struct vectors * vecs)
 		}
 	}
 
-	reply = MALLOC(maxlen);
+	if (snprint_multipath_topology_json(&reply, vecs) < 0)
+		return 1;
 
-	while (again) {
-		if (!reply)
-			return 1;
-
-		c = reply;
-
-		c += snprint_multipath_topology_json(c, maxlen, vecs);
-		again = (c == reply + maxlen);
-
-		REALLOC_REPLY(reply, again, maxlen);
-	}
-	*r = reply;
-	*len = (int)(c - reply);
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -222,28 +153,16 @@ int
 show_map_json (char ** r, int * len, struct multipath * mpp,
 		   struct vectors * vecs)
 {
-	char * c;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	STRBUF_ON_STACK(reply);
 
 	if (update_multipath(vecs, mpp->alias, 0))
 		return 1;
-	reply = MALLOC(maxlen);
 
-	while (again) {
-		if (!reply)
-			return 1;
+	if (snprint_multipath_map_json(&reply, mpp) < 0)
+		return 1;
 
-		c = reply;
-
-		c += snprint_multipath_map_json(c, maxlen, mpp);
-		again = (c == reply + maxlen);
-
-		REALLOC_REPLY(reply, again, maxlen);
-	}
-	*r = reply;
-	*len = (int)(c - reply);
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -414,58 +333,40 @@ cli_list_maps_json (void * v, char ** reply, int * len, void * data)
 int
 cli_list_wildcards (void * v, char ** reply, int * len, void * data)
 {
-	char * c;
+	STRBUF_ON_STACK(buf);
 
-	*reply = MALLOC(INITIAL_REPLY_LEN);
-
-	if (!*reply)
+	if (snprint_wildcards(&buf) < 0)
 		return 1;
 
-	c = *reply;
-	c += snprint_wildcards(c, INITIAL_REPLY_LEN);
-
-	*len = INITIAL_REPLY_LEN;
+	*len = get_strbuf_len(&buf) + 1;
+	*reply = steal_strbuf_str(&buf);
 	return 0;
 }
 
 int
 show_status (char ** r, int *len, struct vectors * vecs)
 {
-	char * c;
-	char * reply;
+	STRBUF_ON_STACK(reply);
 
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	reply = MALLOC(maxlen);
-
-	if (!reply)
+	if (snprint_status(&reply, vecs) < 0)
 		return 1;
 
-	c = reply;
-	c += snprint_status(c, reply + maxlen - c, vecs);
-
-	*r = reply;
-	*len = (int)(c - reply + 1);
+	*len = get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
 int
 show_daemon (char ** r, int *len)
 {
-	char * c;
-	char * reply;
+	STRBUF_ON_STACK(reply);
 
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	reply = MALLOC(maxlen);
-
-	if (!reply)
+	if (print_strbuf(&reply, "pid %d %s\n",
+			 daemon_pid, daemon_status()) < 0)
 		return 1;
 
-	c = reply;
-	c += snprintf(c, INITIAL_REPLY_LEN, "pid %d %s\n",
-		      daemon_pid, daemon_status());
-
-	*r = reply;
-	*len = (int)(c - reply + 1);
+	*len = get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -473,26 +374,13 @@ int
 show_map (char ** r, int *len, struct multipath * mpp, char * style,
 	  int pretty)
 {
-	char * c;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	STRBUF_ON_STACK(reply);
 
-	reply = MALLOC(maxlen);
-	while (again) {
-		if (!reply)
-			return 1;
+	if (snprint_multipath(&reply, style, mpp, pretty) < 0)
+		return 1;
 
-		c = reply;
-		c += snprint_multipath(c, reply + maxlen - c, style,
-				       mpp, pretty);
-
-		again = (c == reply + maxlen - 1);
-
-		REALLOC_REPLY(reply, again, maxlen);
-	}
-	*r = reply;
-	*len = (int)(c - reply + 1);
+	*len = get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -500,51 +388,34 @@ int
 show_maps (char ** r, int *len, struct vectors * vecs, char * style,
 	   int pretty)
 {
+	STRBUF_ON_STACK(reply);
 	int i;
 	struct multipath * mpp;
-	char * c, *header;
-	char * reply;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	int hdr_len = 0;
 
 	get_multipath_layout(vecs->mpvec, 1);
 	foreign_multipath_layout();
 
-	reply = MALLOC(maxlen);
-
-	while (again) {
-		if (!reply)
-			return 1;
-
-		c = reply;
-		if (pretty)
-			c += snprint_multipath_header(c, reply + maxlen - c,
-						      style);
-		header = c;
-
-		vector_foreach_slot(vecs->mpvec, mpp, i) {
-			if (update_multipath(vecs, mpp->alias, 0)) {
-				i--;
-				continue;
-			}
-			c += snprint_multipath(c, reply + maxlen - c,
-					       style, mpp, pretty);
+	if (pretty && (hdr_len = snprint_multipath_header(&reply, style)) < 0)
+		return 1;
 
+	vector_foreach_slot(vecs->mpvec, mpp, i) {
+		if (update_multipath(vecs, mpp->alias, 0)) {
+			i--;
+			continue;
 		}
-		c += snprint_foreign_multipaths(c, reply + maxlen - c,
-						style, pretty);
-		again = (c == reply + maxlen - 1);
-
-		REALLOC_REPLY(reply, again, maxlen);
+		if (snprint_multipath(&reply, style, mpp, pretty) < 0)
+			return 1;
 	}
+	if (snprint_foreign_multipaths(&reply, style, pretty) < 0)
+		return 1;
 
-	if (pretty && c == header) {
+	if (pretty && get_strbuf_len(&reply) == (size_t)hdr_len)
 		/* No output - clear header */
-		*reply = '\0';
-		c = reply;
-	}
-	*r = reply;
-	*len = (int)(c - reply + 1);
+		truncate_strbuf(&reply, 0);
+
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -715,6 +586,15 @@ cli_add_path (void * v, char ** reply, int * len, void * data)
 	pp = find_path_by_dev(vecs->pathvec, param);
 	if (pp && pp->initialized != INIT_REMOVED) {
 		condlog(2, "%s: path already in pathvec", param);
+
+		if (pp->recheck_wwid == RECHECK_WWID_ON &&
+		    check_path_wwid_change(pp)) {
+			condlog(0, "%s: wwid changed. Removing device",
+				pp->dev);
+			handle_path_wwid_change(pp, vecs);
+			return 1;
+		}
+
 		if (pp->mpp)
 			return 0;
 	} else if (pp) {
@@ -743,7 +623,8 @@ cli_add_path (void * v, char ** reply, int * len, void * data)
 				/* Have the checker reinstate this path asap */
 				pp->tick = 1;
 				return 0;
-			} else if (!ev_remove_path(pp, vecs, true))
+			} else if (ev_remove_path(pp, vecs, true) &
+				   REMOVE_PATH_SUCCESS)
 				/* Path removed in ev_remove_path() */
 				pp = NULL;
 			else {
@@ -792,8 +673,7 @@ cli_add_path (void * v, char ** reply, int * len, void * data)
 	}
 	return ev_add_path(pp, vecs, 1);
 blacklisted:
-	*reply = strdup("blacklisted\n");
-	*len = strlen(*reply) + 1;
+	SET_REPLY_AND_LEN(reply, len, "blacklisted\n");
 	condlog(2, "%s: path blacklisted", param);
 	return 0;
 }
@@ -804,6 +684,7 @@ cli_del_path (void * v, char ** reply, int * len, void * data)
 	struct vectors * vecs = (struct vectors *)data;
 	char * param = get_keyparam(v, PATH);
 	struct path *pp;
+	int ret;
 
 	param = convert_dev(param, 1);
 	condlog(2, "%s: remove path (operator)", param);
@@ -812,7 +693,12 @@ cli_del_path (void * v, char ** reply, int * len, void * data)
 		condlog(0, "%s: path already removed", param);
 		return 1;
 	}
-	return ev_remove_path(pp, vecs, 1);
+	ret = ev_remove_path(pp, vecs, 1);
+	if (ret == REMOVE_PATH_DELAY)
+		SET_REPLY_AND_LEN(reply, len, "delayed\n");
+	else if (ret == REMOVE_PATH_MAP_ERROR)
+		SET_REPLY_AND_LEN(reply, len, "map reload error. removed\n");
+	return (ret == REMOVE_PATH_FAILURE);
 }
 
 int
@@ -836,21 +722,21 @@ cli_add_map (void * v, char ** reply, int * len, void * data)
 		invalid = 1;
 	pthread_cleanup_pop(1);
 	if (invalid) {
-		*reply = strdup("blacklisted\n");
-		*len = strlen(*reply) + 1;
+		SET_REPLY_AND_LEN(reply, len, "blacklisted\n");
 		condlog(2, "%s: map blacklisted", param);
 		return 1;
 	}
 	do {
 		if (dm_get_major_minor(param, &major, &minor) < 0)
-			condlog(2, "%s: not a device mapper table", param);
+			condlog(count ? 2 : 3,
+				"%s: not a device mapper table", param);
 		else {
 			sprintf(dev_path, "dm-%d", minor);
 			alias = dm_mapname(major, minor);
 		}
 		/*if there is no mapname found, we first create the device*/
 		if (!alias && !count) {
-			condlog(2, "%s: mapname not found for %d:%d",
+			condlog(3, "%s: mapname not found for %d:%d",
 				param, major, minor);
 			get_refwwid(CMD_NONE, param, DEV_DEVMAP,
 				    vecs->pathvec, &refwwid);
@@ -860,7 +746,6 @@ cli_add_map (void * v, char ** reply, int * len, void * data)
 				    != CP_OK)
 					condlog(2, "%s: coalesce_paths failed",
 									param);
-				dm_lib_release();
 				FREE(refwwid);
 			}
 		} /*we attempt to create device only once*/
@@ -952,12 +837,12 @@ cli_reload(void *v, char **reply, int *len, void *data)
 int resize_map(struct multipath *mpp, unsigned long long size,
 	       struct vectors * vecs)
 {
-	char params[PARAMS_SIZE] = {0};
+	char *params __attribute__((cleanup(cleanup_charp))) = NULL;
 	unsigned long long orig_size = mpp->size;
 
 	mpp->size = size;
 	update_mpp_paths(mpp, vecs->pathvec);
-	if (setup_map(mpp, params, PARAMS_SIZE, vecs) != 0) {
+	if (setup_map(mpp, &params, vecs) != 0) {
 		condlog(0, "%s: failed to setup map for resize : %s",
 			mpp->alias, strerror(errno));
 		mpp->size = orig_size;
@@ -1032,7 +917,6 @@ cli_resize(void *v, char **reply, int *len, void *data)
 	if (resize_map(mpp, size, vecs) != 0)
 		return 1;
 
-	dm_lib_release();
 	if (setup_multipath(vecs, mpp) != 0)
 		return 1;
 	sync_map_state(mpp);
@@ -1196,7 +1080,7 @@ cli_reconfigure(void * v, char ** reply, int * len, void * data)
 
 	condlog(2, "reconfigure (operator)");
 
-	rc = set_config_state(DAEMON_CONFIGURE); 
+	rc = set_config_state(DAEMON_CONFIGURE);
 	if (rc == ETIMEDOUT) {
 		condlog(2, "timeout starting reconfiguration");
 		return 1;
@@ -1347,34 +1231,20 @@ cli_fail(void * v, char ** reply, int * len, void * data)
 int
 show_blacklist (char ** r, int * len)
 {
-	char *c = NULL;
-	char *reply = NULL;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	STRBUF_ON_STACK(reply);
 	struct config *conf;
-	int fail = 0;
-
-	reply = MALLOC(maxlen);
+	bool fail;
 
 	conf = get_multipath_config();
 	pthread_cleanup_push(put_multipath_config, conf);
-	while (again) {
-		if (!reply) {
-			fail = 1;
-			break;
-		}
-
-		c = reply;
-		c += snprint_blacklist_report(conf, c, maxlen);
-		again = (c == reply + maxlen);
-		REALLOC_REPLY(reply, again, maxlen);
-	}
+	fail = snprint_blacklist_report(conf, &reply) < 0;
 	pthread_cleanup_pop(1);
 
 	if (fail)
 		return 1;
-	*r = reply;
-	*len = (int)(c - reply + 1);
+
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 	return 0;
 }
 
@@ -1389,34 +1259,20 @@ cli_list_blacklist (void * v, char ** reply, int * len, void * data)
 int
 show_devices (char ** r, int * len, struct vectors *vecs)
 {
-	char *c = NULL;
-	char *reply = NULL;
-	unsigned int maxlen = INITIAL_REPLY_LEN;
-	int again = 1;
+	STRBUF_ON_STACK(reply);
 	struct config *conf;
-	int fail = 0;
-
-	reply = MALLOC(maxlen);
+	bool fail;
 
 	conf = get_multipath_config();
 	pthread_cleanup_push(put_multipath_config, conf);
-	while (again) {
-		if (!reply) {
-			fail = 1;
-			break;
-		}
-
-		c = reply;
-		c += snprint_devices(conf, c, maxlen, vecs);
-		again = (c == reply + maxlen);
-		REALLOC_REPLY(reply, again, maxlen);
-	}
+	fail = snprint_devices(conf, &reply, vecs) < 0;
 	pthread_cleanup_pop(1);
 
 	if (fail)
 		return 1;
-	*r = reply;
-	*len = (int)(c - reply + 1);
+
+	*len = (int)get_strbuf_len(&reply) + 1;
+	*r = steal_strbuf_str(&reply);
 
 	return 0;
 }
@@ -1521,7 +1377,7 @@ cli_getprkey(void * v, char ** reply, int * len, void * data)
 	struct multipath * mpp;
 	struct vectors * vecs = (struct vectors *)data;
 	char *mapname = get_keyparam(v, MAP);
-	char *flagstr = "";
+	uint64_t key;
 
 	mapname = convert_dev(mapname, 0);
 	condlog(3, "%s: get persistent reservation key (operator)", mapname);
@@ -1534,17 +1390,16 @@ cli_getprkey(void * v, char ** reply, int * len, void * data)
 	if (!*reply)
 		return 1;
 
-	if (!get_be64(mpp->reservation_key)) {
+	key = get_be64(mpp->reservation_key);
+	if (!key) {
 		sprintf(*reply, "none\n");
-		*len = strlen(*reply) + 1;
+		*len = sizeof("none\n");
 		return 0;
 	}
-	if (mpp->sa_flags & MPATH_F_APTPL_MASK)
-		flagstr = ":aptpl";
-	snprintf(*reply, 26, "0x%" PRIx64 "%s\n",
-		 get_be64(mpp->reservation_key), flagstr);
-	(*reply)[19] = '\0';
-	*len = strlen(*reply) + 1;
+
+	/* This snprintf() can't overflow - PRIx64 needs max 16 chars */
+	*len = snprintf(*reply, 26, "0x%" PRIx64 "%s\n", key,
+			mpp->sa_flags & MPATH_F_APTPL_MASK ? ":aptpl" : "") + 1;
 	return 0;
 }
 
